﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using Diplo.Libs.Drawing;
using Diplo.GraphObjects;
using Diplo.Configs;
using Diplo.Libs.Geometry;

namespace Diplo.DrawAlgorithms.Lib
{
    /// <summary>
    /// Pomocna trieda pre A* algoritmus
    /// </summary>
    class AStarNeighbor
    {

        /// <summary>
        /// Pre daný bod nájde susedov. Pre susedov musí platiť niekoľko pravidiel.
        /// </summary>
        /// <param name="g">Komponent, pre kreslenie</param>
        /// <param name="graf">Graf - vrchply, hrany a pod.</param>
        /// <param name="start">Start vrchol, z ktoreho hladame cestu</param>
        /// <param name="end">Posledny vrchol, do ktoreho hladame cestu</param>
        /// <param name="p"></param>
        /// <param name="fieldSize"></param>
        /// <param name="granul">Ako daleko maju byt susedia od seba</param>
        /// <param name="polygon">Polygon, v ktorom sa mozu nachadzat cesty</param>
        /// <returns></returns>
        static public int[,] getNeighborNodes(Graphics g, Graph graf, Point start, Point end, Point p, int fieldSize, int granul, Stack<Point> polygon)
        {
            int[,] neigh = new int[8, 2];
            // inicializacia
            for (int i = 0; i < 8; i++)
            {
                for (int j = 0; j < 2; j++)
                {
                    neigh[i, j] = -1;
                }
            }

            // vrchna randa
            if (isPointAvailible(g, graf, start, end, p.X - granul, p.Y + granul, fieldSize, polygon))
            {
                neigh[0, 0] = p.X - granul;
                neigh[0, 1] = p.Y + granul;
            }

            if (isPointAvailible(g, graf, start, end, p.X, p.Y + granul, fieldSize, polygon))
            {
                neigh[1, 0] = p.X;
                neigh[1, 1] = p.Y + granul;
            }

            if (isPointAvailible(g, graf, start, end, p.X + granul, p.Y + granul, fieldSize, polygon))
            {
                neigh[2, 0] = p.X + granul;
                neigh[2, 1] = p.Y + granul;
            }

            // stredna rada
            if (isPointAvailible(g, graf, start, end, p.X - granul, p.Y, fieldSize, polygon))
            {
                neigh[3, 0] = p.X - granul;
                neigh[3, 1] = p.Y;
            }

            if (isPointAvailible(g, graf, start, end, p.X + granul, p.Y, fieldSize, polygon))
            {
                neigh[4, 0] = p.X + granul;
                neigh[4, 1] = p.Y;
            }

            // dolna rada
            if (isPointAvailible(g, graf, start, end, p.X - granul, p.Y - granul, fieldSize, polygon))
            {
                neigh[5, 0] = p.X - granul;
                neigh[5, 1] = p.Y - granul;
            }

            if (isPointAvailible(g, graf, start, end, p.X, p.Y - granul, fieldSize, polygon))
            {
                neigh[6, 0] = p.X;
                neigh[6, 1] = p.Y - granul;
            }

            if (isPointAvailible(g, graf, start, end, p.X + granul, p.Y - granul, fieldSize, polygon))
            {
                neigh[7, 0] = p.X + granul;
                neigh[7, 1] = p.Y - granul;
            }
            
            return neigh;
        }

        /// <summary>
        /// <para>Overenie, či bod je dobrý sused. Sused musí ležať v poligóne (ak je dodaný poligón), 
        /// musí mať polohu v intervale (0, fieldsize) a cesta by nemala prechádzať cez žiadny bod grafu.</para>
        /// </summary>
        /// <param name="g">Grafický komponent</param>
        /// <param name="graf">Inštancia grafu</param>
        /// <param name="start">ŠTART bod</param>
        /// <param name="end">KONIEC bod</param>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="fieldSize"></param>
        /// <param name="polygon"></param>
        /// <returns></returns>
        static private Boolean isPointAvailible(Graphics g, Graph graf, Point start, Point end, int x, int y, int fieldSize, Stack<Point> polygon)
        {
            Boolean good = true;

            //bod musi lezat v poligone, ak je definovany poligon
            if (polygon.Count > 0)
            {
                if (!PointInPolygon.isPointInPolygon(g, polygon, new System.Drawing.Point(x, y)))
                {
                    // hackfix - ak je velmi blizko ciela, je to OK
                    int distance = TwoPoints.manhatan_distance(start, new System.Drawing.Point(x, y));
                    if (distance < 25)
                    {
                        return true;
                    }

                    return false;
                }
            }
            
            //else {  }

            // nie je mimo x-osi plochy
            if (x < 0 || x > fieldSize)
            {
                good = false;
            }

            if (y < 0 || y > fieldSize)
            {
                good = false;
            }

            if (nodeDetected(g, graf, start, end, x, y))
            {
                good = false;
            }

            return good;
        }

        /// <summary>
        /// <para>Metóda určí, či bod leží v súkromnej oblasti niektoré vrcholu grafu. Súkromna oblasť je 
        /// štvorcový rámec vrcholu (v ktorom nemôže ležať bod cesty ... ).</para>
        /// </summary>
        /// <param name="g"></param>
        /// <param name="graf"></param>
        /// <param name="start"></param>
        /// <param name="end"></param>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <returns></returns>
        static private bool nodeDetected(Graphics g, Graph graf, Point start, Point end, int x, int y)
        {
            foreach (Vertex v in graf.vertices)
            {
                int x1 = (int)Math.Round(v.pos[0]);
                int y1 = (int)Math.Round(v.pos[1]);

                if ((x1 + v.size / 2 == start.X && y1 + v.size / 2 == start.Y) || (x1 + v.size / 2 == end.X && y1 + v.size / 2 == end.Y)) { }
                else
                {
                    if (x == x1 && y == y1) { }
                    else
                    {
                        int left = x1 - Config.vertexArea;
                        int up = y1 - Config.vertexArea;
                        int right = x1 + v.size + Config.vertexArea;
                        int down = y1 + v.size + Config.vertexArea;

                        if ((left < x && right > x) && (up < y && down > y))
                        {
                            return true;
                        }

                    }
                }

            }

            return false;
        }

    }
}
