﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using Diplo.Configs;

namespace Diplo.Libs.Geometry
{
    /// <summary>
    /// Výpočet rôznych vzťahov medzi úsečkou a bodom.
    /// </summary>
    class PointLine
    {
        /// <summary>
        /// <para>Výpočet bodu, ktorý leží na úsečke v ceste, pričom je od jedného bodu (špecifikovaný
        /// ako parameter) vzdialený o špecifickú vzdialenosť.</para>
        /// </summary>
        /// <param name="g">Grafický komponent.</param>
        /// <param name="path">Cesta</param>
        /// <param name="afterMiddlePoint">Pred ktorým bodom sa má hľadať stred.</param>
        /// <param name="distance">Vzdialenosť v ktorej má byť nový bod.</param>
        /// <returns>Nájdený bod.</returns>
        public static Point computePointOnLineSegmentInDistance(Graphics g, Stack<Point> path, int afterMiddlePoint, double distance)
        {
            // oznacenie dvoch vrcholov, medzi ktorymi budeme hladat stred
            Point p1 = path.ElementAt(afterMiddlePoint - 1);
            Point p2 = path.ElementAt(afterMiddlePoint);

            // najdeme stredny bod
            Point midle = LineSegment.middlePoint(p1, p2, distance);
            if (Config.showMiddlePoints)
            {
                g.FillEllipse(new SolidBrush(Color.Green), midle.X - 5, midle.Y - 5, 10, 10);
            }

            return midle;
        }

        /// <summary>
        /// <para>Pre dany bod a cestu (mnozinu usecok), vypocita najkratsiu moznu vzdialenost bodu od usecky.</para>
        /// <para>Pre kazdu usecku cesty sa vypocita vzdialenost bodu od usecky a nakoniec sa vyberie minimum.</para>
        /// </summary>
        /// <param name="g"></param>
        /// <param name="middlePoint"></param>
        /// <param name="oldPath"></param>
        /// <returns></returns>
        public static int computePointToPathDistance(Graphics g, Point middlePoint, Stack<Point> oldPath)
        {
            double min = -1;
            int best = -1;
            // pre kazdy usek vypocitame vzdialenost
            for (int i = 1; i < oldPath.Count(); i++)
            {
                double[] PointA = new double[2];
                double[] PointB = new double[2];
                double[] PointC = new double[2];

                PointA[0] = oldPath.ElementAt(i - 1).X;
                PointA[1] = oldPath.ElementAt(i - 1).Y;

                PointB[0] = oldPath.ElementAt(i).X;
                PointB[1] = oldPath.ElementAt(i).Y;

                PointC[0] = middlePoint.X;
                PointC[1] = middlePoint.Y;

                double res = PointLine.LineToPointDistance2D(PointA, PointB, PointC, true);
                if (min == -1 || res < min)
                {
                    best = i;
                    min = Math.Floor(res);
                }
            }

            try
            {
                return Convert.ToInt32(min);
            }
            catch (Exception e)
            {
                return 0;
            }
        }

        /// <summary>
        /// <para>Metóda na výpočet vzdialenosť bodu C od úsečky AB / priamky na ktorej ležia body A, B</para>
        /// </summary>
        /// <param name="pointA">Bod A úsečky / priamky</param>
        /// <param name="pointB">Bod B úsečky / priamky</param>
        /// <param name="pointC">Bod C, pre ktorý sa zisťuje vzdialenosť od úsečky / priamky</param>
        /// <param name="isSegment">Ak je TRUE, AB je úsečka. Ináč priamka.</param>
        /// <returns></returns>
        public static double LineToPointDistance2D(double[] pointA, double[] pointB, double[] pointC, bool isSegment)
        {
            double dist = CrossProduct(pointA, pointB, pointC) / TwoPoints.euclidDistance(pointA, pointB);
            if (isSegment)
            {
                double dot1 = DotProduct(pointA, pointB, pointC);
                if (dot1 > 0)
                    return TwoPoints.euclidDistance(pointB, pointC);

                double dot2 = DotProduct(pointB, pointA, pointC);
                if (dot2 > 0)
                    return TwoPoints.euclidDistance(pointA, pointC);
            }
            return Math.Abs(dist);
        }

        /// <summary>
        /// <para>Výpočet : dot product AB . AC</para>
        /// </summary>
        /// <param name="pointA">Bod A</param>
        /// <param name="pointB">Bod B</param>
        /// <param name="pointC">Bod C</param>
        /// <returns></returns>
        internal static double DotProduct(double[] pointA, double[] pointB, double[] pointC)
        {
            double[] AB = new double[2];
            double[] BC = new double[2];
            AB[0] = pointB[0] - pointA[0];
            AB[1] = pointB[1] - pointA[1];
            BC[0] = pointC[0] - pointB[0];
            BC[1] = pointC[1] - pointB[1];
            double dot = AB[0] * BC[0] + AB[1] * BC[1];

            return dot;
        }

        /// <summary>
        /// <para>Cross product AB x AC</para>
        /// </summary>
        /// <param name="pointA">Bod A</param>
        /// <param name="pointB">Bod B</param>
        /// <param name="pointC">Bod C</param>
        /// <returns></returns>
        public static double CrossProduct(double[] pointA, double[] pointB, double[] pointC)
        {
            double[] AB = new double[2];
            double[] AC = new double[2];
            AB[0] = pointB[0] - pointA[0];
            AB[1] = pointB[1] - pointA[1];
            AC[0] = pointC[0] - pointA[0];
            AC[1] = pointC[1] - pointA[1];
            double cross = AB[0] * AC[1] - AB[1] * AC[0];

            return cross;
        }
    }
}
